!
! Module: imcrop_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: The task clips out a subregion of an image and returns this. WCS keywords, if present, are adjusted correctly.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module imcrop_mod

  use dal
  use cal
  use param  
  use errorHandling

  use dal_aux
  use wcs_aux
  use array_utils

  implicit none 

  private 

  ! Command-line parameters:
  character(256) ::&
    inImageSetName='',&
    outImageSetName=''
  character(20) ::&
    cropStyle=''
  real(single) ::&
    xClipFractionLo=0.0,&
    xClipFractionHi=0.0,&
    yClipFractionLo=0.0,&
    yClipFractionHi=0.0
  integer(int16) ::&
    xClipLo=0,&
    xClipHi=0,&
    yClipLo=0,&
    yClipHi=0
  logical ::&
!    withXYclip,&
    clipBlankBorder=.false.

  public imcrop

contains 

  !---------------------------------------------------------------------
  subroutine imcrop
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: inSet, outSet
    type(ArrayT) :: inArray, outArray
    integer(int32), pointer :: inDimVector(:), outDimVector(:)
    integer(int16) :: xsize, ysize, minx, maxx, miny, maxy, minx_temp, miny_temp
    integer :: status
    integer(int32) :: arrayDataType
    type(WcsT) :: inWcs, outWcs

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    inSet = dataSet(inImageSetName, READ)
    inArray = array(inSet, 0)
    arrayDataType = dataType(inArray)
    inDimVector => dimensions(inArray)
    xsize = inDimVector(0)
    ysize = inDimVector(1)

    select case(cropStyle)
    case('fraction')
      if (clipBlankBorder) then
        call findExtentBlankBorder(inArray, minx, maxx, miny, maxy, status)
if (status .eq. -1) call error('imageAllZero',"You have set --trimborder=yes, &
&but imcrop can't trim away the zero-valued bits as you request, because &
&there would be nothing left!")

if (status .eq. -2) call error('sizelessImageArray',"The dimensions of the &
&input image array are zero.")

      else
        minx = 1
        maxx = xsize
        miny = 1
        maxy = ysize
      end if

      ! Because in copyDataBetweenArrays the arrays start at 0:
      minx = minx - 1
      maxx = maxx - 1
      miny = miny - 1
      maxy = maxy - 1

!      if (withXYclip) then
        minx_temp = nint(minx * (1.0 - xClipFractionLo)&
          + maxx * xClipFractionLo)
        maxx = nint(minx * (1.0 - xClipFractionHi) + maxx * xClipFractionHi)
        minx = minx_temp
        miny_temp = nint(miny * (1.0 - yClipFractionLo)&
          + maxy * yClipFractionLo)
        maxy = nint(miny * (1.0 - yClipFractionHi) + maxy * yClipFractionHi)
        miny = miny_temp
if (minx >= maxx) call error('badXclipValues',"--xfraclo must be less &
&than --xfrachi.")

if (miny >= maxy) call error('badYclipValues',"--yfraclo must be less &
&than --yfrachi.")
!      end if

    case('pixel')
      minx = xClipLo
      maxx = xClipHi
      miny = yClipLo
      maxy = yClipHi

      if (maxx > xsize) then
        call warning('xhiTooLarge',"")
        maxx = xsize
      end if

      if (maxy > ysize) then
        call warning('yhiTooLarge',"")
        maxy = ysize
      end if

      ! Because in copyDataBetweenArrays the arrays start at 0:
      minx = minx - 1
      maxx = maxx - 1
      miny = miny - 1
      maxy = maxy - 1

    case default
call error('',"")
    end select

    allocate(outDimVector(0:1))
    outDimVector(0) = 1 + maxx - minx
    outDimVector(1) = 1 + maxy - miny

    outSet = dataSet(outImageSetName, CREATE)
    call setAttributes(attributable(outSet), attributable(inSet))
    outArray = addArray(outSet, 'PRIMARY', dataType(inArray), outDimVector)

    call copyDataBetweenArrays(inArray, outArray, minx, maxx, miny, maxy)
    call setAttributes(attributable(outArray), attributable(inArray))

    ! Fix up new wcs:

    call readWcs(inSet, inArray, inWcs) ! in ssclib/wcs_aux
    call release(inArray)
    call release(inSet)
    outWcs = inWcs
    outWcs%x%refPixels = outWcs%x%refPixels - minx
    outWcs%y%refPixels = outWcs%y%refPixels - miny

    call writeWcs(outArray, outWcs) ! in ssclib/wcs_aux
    call release(outArray)
    call release(outSet)

  end subroutine imcrop

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    inImageSetName = stringParameter('inset')
    outImageSetName = stringParameter('outset')
    cropStyle = stringParameter('cropstyle')
    select case(cropStyle)
    case('fraction')
      clipBlankBorder = booleanParameter('trimborder')
!      withXYclip = booleanParameter('withxyclip')
      xClipFractionLo = realParameter('xfraclo')
      xClipFractionHi = realParameter('xfrachi')
      yClipFractionLo = realParameter('yfraclo')
      yClipFractionHi = realParameter('yfrachi')

    case('pixel')
      xClipLo = intParameter('xlo')
      xClipHi = intParameter('xhi')
      yClipLo = intParameter('ylo')
      yClipHi = intParameter('yhi')

    case default
call error('',"")
    end select

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine findExtentBlankBorder(inArray, minx, maxx, miny, maxy, status)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT),   intent(in)  :: inArray
    integer(int16), intent(out) :: minx, maxx, miny, maxy
    integer,        intent(out) :: status

    integer(int8),  pointer :: arrayDataInt8(:,:)
    integer(int16), pointer :: arrayDataInt16(:,:)
    integer(int32), pointer :: arrayDataInt32(:,:)
    real(single),   pointer :: arrayDataReal32(:,:)
    real(double),   pointer :: arrayDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start

    select case(dataType(inArray))
    case(INTEGER8)
      arrayDataInt8 => int8Array2Data(inArray)
!      call maskArrayLimits(logical(arrayDataInt8 .ne. 0, bool)&
      call maskArrayLimits((arrayDataInt8 .ne. 0)&
        , minx, maxx, miny, maxy, status)
    case(INTEGER16)
      arrayDataInt16 => int16Array2Data(inArray)
      call maskArrayLimits((arrayDataInt16 .ne. 0), minx, maxx, miny, maxy&
        , status)
!      call maskArrayLimits(logical(arrayDataInt16 .ne. 0, bool)&
!        , minx, maxx, miny, maxy, status)
    case(INTEGER32)
      arrayDataInt32 => int32Array2Data(inArray)
      call maskArrayLimits((arrayDataInt32 .ne. 0), minx, maxx, miny, maxy&
        , status)
!      call maskArrayLimits(logical(arrayDataInt32 .ne. 0, bool)&
!        , minx, maxx, miny, maxy, status)
    case(REAL32)
      arrayDataReal32 => real32Array2Data(inArray)
      call maskArrayLimits((arrayDataReal32 .ne. 0.0), minx, maxx, miny, maxy&
        , status)
!      call maskArrayLimits(logical(arrayDataReal32 .ne. 0.0, bool)&
!        , minx, maxx, miny, maxy, status)
    case(REAL64)
      arrayDataReal64 => real64Array2Data(inArray)
      call maskArrayLimits((arrayDataReal64 .ne. 0.0), minx, maxx, miny, maxy&
        , status)
!      call maskArrayLimits(logical(arrayDataReal64 .ne. 0.0, bool)&
!        , minx, maxx, miny, maxy, status)
!    case(BOOLEAN)
!      arrayDataBool => boolArray2Data(inArray)
!      call maskArrayLimits(arrayDataBool, minx, maxx, miny, maxy&
!        , status)
    case default
call error('badArrayDataType',"The data type of the input image was not &
&recognized.")
    end select

  end subroutine findExtentBlankBorder

  !---------------------------------------------------------------------
  subroutine copyDataBetweenArrays(fromArray, toArray&
  , startX, finisX, startY, finisY)
  !---------------------------------------------------------------------


!*** put in dal_aux????



    implicit none

    type(ArrayT), intent(in) :: fromArray, toArray
    integer(int16), intent(in) :: startX, finisX, startY, finisY

    integer(int8),  pointer :: arrayDataInt8From(:,:),   arrayDataInt8To(:,:)
    integer(int16), pointer :: arrayDataInt16From(:,:),  arrayDataInt16To(:,:)
    integer(int32), pointer :: arrayDataInt32From(:,:),  arrayDataInt32To(:,:)
    real(single),   pointer :: arrayDataReal32From(:,:), arrayDataReal32To(:,:)
    real(double),   pointer :: arrayDataReal64From(:,:), arrayDataReal64To(:,:)

    !-------------------------------------------------------------------
    ! Start

    select case(dataType(fromArray))
    case(INTEGER8)
      arrayDataInt8From => int8Array2Data(fromArray)
      arrayDataInt8To   => int8Array2Data(toArray)
      arrayDataInt8To = arrayDataInt8From(startX:finisX,startY:finisY)
    case(INTEGER16)
      arrayDataInt16From => int16Array2Data(fromArray)
      arrayDataInt16To   => int16Array2Data(toArray)
      arrayDataInt16To = arrayDataInt16From(startX:finisX,startY:finisY)
    case(INTEGER32)
      arrayDataInt32From => int32Array2Data(fromArray)
      arrayDataInt32To   => int32Array2Data(toArray)
      arrayDataInt32To = arrayDataInt32From(startX:finisX,startY:finisY)
    case(REAL32)
      arrayDataReal32From => real32Array2Data(fromArray)
      arrayDataReal32To   => real32Array2Data(toArray)
      arrayDataReal32To = arrayDataReal32From(startX:finisX,startY:finisY)
    case(REAL64)
      arrayDataReal64From => real64Array2Data(fromArray)
      arrayDataReal64To   => real64Array2Data(toArray)
      arrayDataReal64To = arrayDataReal64From(startX:finisX,startY:finisY)
!    case(BOOLEAN)
!      call copyColsTweenBool(fromCol, fromColStartRow, numFromRows&
!        , toCol, toColStartRow)
    case default
call error('badArrayDataType', "The data type of the input image was not &
&recognized.")
    end select

  end subroutine copyDataBetweenArrays

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module imcrop_mod
