!+
! Module: imweightadd_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Makes a weighted sum of input images.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module imweightadd_mod

  use dal
  use param  
  use errorHandling

!  use array_utils	! for valueAtHistoFraction
  use dal_aux		! for readArrayData
!  use math_utils	! for incompleteGammaP
!  use minimizations	! for simplexMinimization, riddersMethod
  use save_image
!  use wcs_aux		! for readWcs, writeWcs

  implicit none 

  private 

  ! Command-line parameters:
  character(256), allocatable ::&
    inImageSetNames(:)
  character(256) ::&
    weightSetTabName='',&
    outImageSetName = '',&
    tempImageSetName=''
  character(20) ::&
    weightStyle=''
  real(single), allocatable ::&
    userWeights(:)
  logical ::&
    withWeights = .false.

 public imweightadd

contains 

  !---------------------------------------------------------------------
  subroutine imweightadd
  !---------------------------------------------------------------------

    implicit none

    integer :: numImages
    real(single), allocatable :: weights(:)

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    numImages = size(inImageSetNames)
    allocate(weights(numImages))
    if (withWeights) then
      select case(weightStyle)
      case('user')
        weights = userWeights
      case('table')
        call readWeights(weights)
      case default
call error('unrecognizedWeightStyle',"")
      end select

    else
      weights = 1.0
    end if

    call addAndSaveImages(inImageSetNames, weights)

  end subroutine imweightadd

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer(int16) :: numImages, i

    !-------------------------------------------------------------------
    ! Start

    numImages = parameterCount('imagesets')
    if (numImages <= 0)&
call error('tooFewImages',"")

    allocate(inImageSetNames(numImages))
    do i = 1, numImages
      inImageSetNames(i) = stringParameter('imagesets',  i-1)
    end do

    withWeights = booleanParameter('withweights')
    if (withWeights) then
      weightStyle = stringParameter('weightstyle')
      select case(weightStyle)
      case('user')
        if (numImages .ne. parameterCount('weights'))&
call error('badNumWeights',"")

        allocate(userWeights(numImages))
        do i = 1, numImages
          userWeights(i) = realParameter('weights',  i-1)
        end do

      case('table')
        weightSetTabName = stringParameter('weighttab')
      case default
call error('unrecognizedWeightStyle',"")
      end select
    end if

    outImageSetName  = stringParameter('outimageset')
    tempImageSetName = stringParameter('tempset')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine readWeights(weights)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(out) :: weights(:)

    character(256) :: weightSetName
    character(20) :: weightTabName
    type(DataSetT) :: set
    type(TableT) :: tab

    !-------------------------------------------------------------------
    ! Start

    call splitSetTabName(weightSetTabName, weightSetName, weightTabName)
    ! in ssclib/dal_aux.

    set = dataSet(weightSetName, READ)
    tab = table(set, weightTabName)
if (numberOfRows(tab) .ne. size(weights)) call error('badNumRowsWeightTable',"")
    call readColDataToFixed(tab, 'WEIGHT', weights) ! in ssclib/dal_aux.
    call release(tab)
    call release(set)

  end subroutine readWeights

  !---------------------------------------------------------------------
  subroutine addAndSaveImages(imageSetNames, weights)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetNames(:)
    real(single), intent(in) :: weights(size(imageSetNames))

    integer :: numImages, imageNum
    type(DataSetT) :: set
    type(ArrayT) :: inArray
    integer(int32), allocatable :: imageDataTypes(:)

    !-------------------------------------------------------------------
    ! Start

    numImages = size(imageSetNames)

    ! Decide on data type of output:

    allocate(imageDataTypes(numImages))
    do imageNum = 1, numImages
      set = dataSet(imageSetNames(imageNum), READ)
      inArray = array(set, 0)
      imageDataTypes(imageNum) = dataType(inArray)
if (imageDataTypes(imageNum) == STRING) call error('stringDataType',"")
      call release(set)
    end do

    if (any(imageDataTypes == REAL64)) then
      call addAndSaveImagesReal64(imageSetNames, weights)
    else if (any(imageDataTypes == REAL32)) then
      call addAndSaveImagesReal32(imageSetNames, weights)
    else if (any(imageDataTypes == INTEGER32)) then
      if (any(weights .ne. nint(weights))) call warning('innaccurateSum',"")
      call addAndSaveImagesInteger32(imageSetNames, weights)
    else if (any(imageDataTypes == INTEGER16)) then
      if (any(weights .ne. nint(weights))) call warning('innaccurateSum',"")
      call addAndSaveImagesInteger16(imageSetNames, weights)
    else if (any(imageDataTypes == INTEGER8)) then
      if (any(weights .ne. nint(weights))) call warning('innaccurateSum',"")
      call addAndSaveImagesInteger8(imageSetNames, weights)
!    else if (all(imageDataTypes == BOOLEAN)) then
!      call addAndSaveImagesBoolean(imageSetNames)
    else
call error('unsupportedDataType',"")
    end if

  end subroutine addAndSaveImages

  !---------------------------------------------------------------------
  subroutine addAndSaveImagesReal32(imageSetNames, weights)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetNames(:)
    real(single), intent(in) :: weights(size(imageSetNames))

    integer :: numImages, xSize, ySize, imageNum
    real(single), pointer :: tempImage(:,:), sumImage(:,:)

    !-------------------------------------------------------------------
    ! Start

    numImages = size(imageSetNames)
    call readArrayData(imageSetNames(1), tempImage) ! in dal_aux
    xSize = size(tempImage, 1)
    ySize = size(tempImage, 2)
    allocate(sumImage(xSize, ySize))
    sumImage = tempImage * weights(1)

    deallocate(tempImage)
    do imageNum = 2, numImages
      call readArrayData(imageSetNames(imageNum), tempImage) ! in dal_aux
      if (size(tempImage,1) .ne. xSize .or. size(tempImage,2) .ne. ySize)&
call error('imagesDifferentDimensions',"")
      sumImage = sumImage + tempImage * weights(imageNum)
      deallocate(tempImage)
    end do

    call saveImage(sumImage, imageSetNames(1), outImageSetName&
      , tempImageSetName) ! in ssclib/save_image.

    deallocate(sumImage)

  end subroutine addAndSaveImagesReal32

  !---------------------------------------------------------------------
  subroutine addAndSaveImagesReal64(imageSetNames, weights)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetNames(:)
    real(single), intent(in) :: weights(size(imageSetNames))

    integer :: numImages, xSize, ySize, imageNum
    real(double), pointer :: tempImage(:,:), sumImage(:,:)

    !-------------------------------------------------------------------
    ! Start

    numImages = size(imageSetNames)
    call readArrayData(imageSetNames(1), tempImage) ! in dal_aux
    xSize = size(tempImage, 1)
    ySize = size(tempImage, 2)
    allocate(sumImage(xSize, ySize))
    sumImage = tempImage * real(weights(1), double)

    deallocate(tempImage)
    do imageNum = 2, numImages
      call readArrayData(imageSetNames(imageNum), tempImage) ! in dal_aux
      if (size(tempImage,1) .ne. xSize .or. size(tempImage,2) .ne. ySize)&
call error('imagesDifferentDimensions',"")
      sumImage = sumImage + tempImage * real(weights(imageNum), double)
      deallocate(tempImage)
    end do

    call saveImage(sumImage, imageSetNames(1), outImageSetName&
      , tempImageSetName) ! in ssclib/save_image.

    deallocate(sumImage)

  end subroutine addAndSaveImagesReal64

  !---------------------------------------------------------------------
  subroutine addAndSaveImagesInteger8(imageSetNames, weights)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetNames(:)
    real(single), intent(in) :: weights(size(imageSetNames))

    integer :: numImages, xSize, ySize, imageNum
    integer(int8), pointer :: tempImage(:,:), sumImage(:,:)

    !-------------------------------------------------------------------
    ! Start

    numImages = size(imageSetNames)
    call readArrayData(imageSetNames(1), tempImage) ! in dal_aux
    xSize = size(tempImage, 1)
    ySize = size(tempImage, 2)
    allocate(sumImage(xSize, ySize))
    sumImage = nint(tempImage * weights(1), int8)

    deallocate(tempImage)
    do imageNum = 2, numImages
      call readArrayData(imageSetNames(imageNum), tempImage) ! in dal_aux
      if (size(tempImage,1) .ne. xSize .or. size(tempImage,2) .ne. ySize)&
call error('imagesDifferentDimensions',"")
      sumImage = sumImage + nint(tempImage * weights(imageNum), int8)
      deallocate(tempImage)
    end do

    call saveImage(sumImage, imageSetNames(1), outImageSetName&
      , tempImageSetName) ! in ssclib/save_image.

    deallocate(sumImage)

  end subroutine addAndSaveImagesInteger8

  !---------------------------------------------------------------------
  subroutine addAndSaveImagesInteger16(imageSetNames, weights)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetNames(:)
    real(single), intent(in) :: weights(size(imageSetNames))

    integer :: numImages, xSize, ySize, imageNum
    integer(int16), pointer :: tempImage(:,:), sumImage(:,:)

    !-------------------------------------------------------------------
    ! Start

    numImages = size(imageSetNames)
    call readArrayData(imageSetNames(1), tempImage) ! in dal_aux
    xSize = size(tempImage, 1)
    ySize = size(tempImage, 2)
    allocate(sumImage(xSize, ySize))
    sumImage = nint(tempImage * weights(1), int16)

    deallocate(tempImage)
    do imageNum = 2, numImages
      call readArrayData(imageSetNames(imageNum), tempImage) ! in dal_aux
      if (size(tempImage,1) .ne. xSize .or. size(tempImage,2) .ne. ySize)&
call error('imagesDifferentDimensions',"")
      sumImage = sumImage + nint(tempImage * weights(imageNum), int16)
      deallocate(tempImage)
    end do

    call saveImage(sumImage, imageSetNames(1), outImageSetName&
      , tempImageSetName) ! in ssclib/save_image.

    deallocate(sumImage)

  end subroutine addAndSaveImagesInteger16

  !---------------------------------------------------------------------
  subroutine addAndSaveImagesInteger32(imageSetNames, weights)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetNames(:)
    real(single), intent(in) :: weights(size(imageSetNames))

    integer :: numImages, xSize, ySize, imageNum
    integer(int32), pointer :: tempImage(:,:), sumImage(:,:)

    !-------------------------------------------------------------------
    ! Start

    numImages = size(imageSetNames)
    call readArrayData(imageSetNames(1), tempImage) ! in dal_aux
    xSize = size(tempImage, 1)
    ySize = size(tempImage, 2)
    allocate(sumImage(xSize, ySize))
    sumImage = nint(tempImage * weights(1), int32)

    deallocate(tempImage)
    do imageNum = 2, numImages
      call readArrayData(imageSetNames(imageNum), tempImage) ! in dal_aux
      if (size(tempImage,1) .ne. xSize .or. size(tempImage,2) .ne. ySize)&
call error('imagesDifferentDimensions',"")
      sumImage = sumImage + nint(tempImage * weights(imageNum), int32)
      deallocate(tempImage)
    end do

    call saveImage(sumImage, imageSetNames(1), outImageSetName&
      , tempImageSetName) ! in ssclib/save_image.

    deallocate(sumImage)

  end subroutine addAndSaveImagesInteger32

  !---------------------------------------------------------------------
  subroutine addAndSaveImagesBoolean(imageSetNames)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetNames(:)

    integer :: numImages, xSize, ySize, imageNum
!    logical(bool), pointer :: tempImage(:,:), sumImage(:,:)
    logical, pointer :: tempImage(:,:), sumImage(:,:)

    !-------------------------------------------------------------------
    ! Start

    numImages = size(imageSetNames)
    call readArrayData(imageSetNames(1), tempImage) ! in dal_aux
    xSize = size(tempImage, 1)
    ySize = size(tempImage, 2)
    allocate(sumImage(xSize, ySize))
    sumImage = tempImage

    deallocate(tempImage)
    do imageNum = 2, numImages
      call readArrayData(imageSetNames(imageNum), tempImage) ! in dal_aux
      if (size(tempImage,1) .ne. xSize .or. size(tempImage,2) .ne. ySize)&
call error('imagesDifferentDimensions',"")
      sumImage = sumImage .or. tempImage
      deallocate(tempImage)
    end do

    call saveImage(sumImage, imageSetNames(1), outImageSetName&
      , tempImageSetName) ! in ssclib/save_image.

    deallocate(sumImage)

  end subroutine addAndSaveImagesBoolean

  !---------------------------------------------------------------------
  logical function imagesAreSameDimensions(imageSetNames, xSizeArg, ySizeArg)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in)           :: imageSetNames(:)
    integer(int16), intent(in), optional :: xSizeArg, ySizeArg

    integer :: numImages, xSize, ySize, startNum, imageNum
    real(double), pointer :: tempImage(:,:)

    !-------------------------------------------------------------------
    ! Start

    imagesAreSameDimensions = .false. ! default

    numImages = size(imageSetNames)

    if (present(xSizeArg) .and. present(ySizeArg)) then
      xSize = xSizeArg
      ySize = ySizeArg
      startNum = 1
    else
      call readArrayData(imageSetNames(1), tempImage) ! in dal_aux
      xSize = size(tempImage, 1)
      ySize = size(tempImage, 2)
      deallocate(tempImage)
      startNum = 2
    end if

    do imageNum = startNum, numImages
      call readArrayData(imageSetNames(imageNum), tempImage) ! in dal_aux
      if (size(tempImage,1) .ne. xSize .or. size(tempImage,2) .ne. ySize) then
        deallocate(tempImage)
  return
      end if

      deallocate(tempImage)
    end do

    imagesAreSameDimensions = .true.

  end function imagesAreSameDimensions

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module imweightadd_mod
