//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	tcsadd	--- Add constant offset to TCS UTC data
//
//	DESCRIPTION:
//
//	USAGE:
//		tcsadd [clonetcs=yes|no] [deltat=<val>]
//
//	CHANGE HISTORY:
//		UL, 2003-05-14: 1.0.0: - initial version
//
//	AUTHOR:
//		U. Lammers, Wed May 14 18:15:39 MEST 2003
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/tools/src/tcsfix/tcsadd.cc,v 1.1 2003/05/14 16:40:59 ulammers Exp $
//
#include "DataFilter.h"
#include "OalServer.h"
#include "CalOalUtility.h"
#include "STime.h"

#include "tcsadd.h"

using namespace std;

class GSTime {
public:
	GSTime(int16 day, int32 milli, int16 micro) :
			_day(day), _milli(milli), _micro(micro) {}

	int16 day() const { return _day; }
	int32 milli() const { return _milli; }
	int16 micro() const { return _micro; }

	GSTime &operator-=(double delta) {
		assert(delta>=0.);

		unsigned udelta = UNT(1000*1000*delta+.5);

		unsigned microsec = udelta%1000;
		udelta /= 1000;
		unsigned millisec = udelta%86400000;

		_micro -= microsec;
		_milli -= millisec;
		if (_micro<0) {
			_micro += 1000;
			--_milli;
		}
		if (_milli<0) {
			_milli += 1000*86400;
			--_day;
			assert(_day>=0);	//	cannot get negative
		}

		return *this;
	} // operator-=

	GSTime &operator+=(double delta) {
		assert(delta>=0.);

		unsigned udelta = UNT(1000*1000*delta+.5);

		unsigned microsec = udelta%1000;
		udelta /= 1000;
		unsigned millisec = udelta%86400000;

		_micro += microsec;
		_milli += millisec;
		if (_micro>=1000) {
			_micro -= 1000;
			++_milli;
		}
		if (_milli>=1000*86400) {
			_milli -= 1000*86400;
			++_day;
		}

		return *this;
	} // operator+=

private:
	int16	_day;
	int32	_milli;
	int16	_micro;
};

class TcsFixer {
public:
	TcsFixer(Table *tcstab) : _tcstab(tcstab), _fixedCode(0) {}

	enum { OBDDEL=1<<3 };

	void correct(int fixCode, double offset) {
		if (fixCode & OBDDEL) {
			errstr<<"UTC offsets have been corrected in a "
				    "previous run already"
				  <<Sas::message(Msg::App, Msg::Always);
			return;
		}

		clean();

		unsigned N = _tcstab->rows();

		ColumnData *cdutcday = _tcstab->column("UTCDAY")->data(),
				   *cdutcmilli = _tcstab->column("UTCMILLI")->data(),
				   *cdutcmicro = _tcstab->column("UTCMICRO")->data(),
				   *cdert = _tcstab->column("ERTIME")->data(),
				   *cdertraw = _tcstab->column("ERTIMRAW")->data();
		int16	*utcday		= cdutcday->int16Data();
		int32	*utcmilli	= cdutcmilli->int32Data();
		int16	*utcmicro	= cdutcmicro->int16Data();

		for (unsigned n=0; n<N; ++n) {
			STime ert(CalOalUtility::stringColumnRow(cdert, n)),
				  ertraw(CalOalUtility::stringColumnRow(cdertraw, n));
			GSTime gst(utcday[n], utcmilli[n], utcmicro[n]);
			gst += offset;
			utcday[n] = gst.day();
			utcmilli[n] = gst.milli();
			utcmicro[n] = gst.micro();

			ert += offset;
			ertraw += offset;
			CalOalUtility::stringColumnRowSet(cdert, n, ert.toString(6));
			CalOalUtility::stringColumnRowSet(cdertraw, n,
													  ertraw.toString(6));
		}

		delete cdutcday;
		delete cdutcmilli;
		delete cdutcmicro;
		delete cdert;
		delete cdertraw;

		_fixedCode += UNT(OBDDEL);

		errstr<<"Wrong UTC offset values successfully corrected in `"
			  <<_tcstab->qualifiedName()<<"'"
			  <<Sas::message(Msg::App, Msg::Always);
	} // correct()

	int fixedCode() const { return _fixedCode; }

	//
	//	remove rows with UTCDAY==0
	//	occasionally seen in the early ODFs (e.g. rev78)
	//
	void clean() {
		DataFilter filter("UTCDAY!=0");
		filter.doFilter(*_tcstab);
	} // clean()

	~TcsFixer() {}

private:
	Table		*_tcstab;
	int			_fixedCode;
};

void tcsadd()
{
	const char *version = "1.0.0";				//	version

	//	locate TCS
	const Odf *odf = oalServer.odf();
	string tcsname = odf->selectSet(Xmm::SC, Xmm::TIME_CORRELATION, 1)->name();
	Table *tcstab = dataSetServer->open(tcsname, Dal::Modify)->table(0);

	if (booleanParameter("clonetcs"))
		dataSetServer->copy(tcsname, tcsname + ".old");

	int fixCode = 0;
	Attribute *tcsfixcode;
	const string TCSFIXC = "TCSFIXC";
	if (tcstab->hasAttribute(TCSFIXC)) {
		tcsfixcode = tcstab->attribute(TCSFIXC);
		fixCode = tcsfixcode->asInt();
	} else
		tcsfixcode = tcstab->addAttribute(TCSFIXC,
								   "Numerical code of applied corrections");

	TcsFixer tcsfixer(tcstab);
	tcsfixer.correct(fixCode, realParameter("deltat"));

	//	add TCSFIXED
	const string TCSFIXED = "TCSFIXED",
				 TCSFIXV  = "TCSFIXV";
	Attribute
		*tcsfixed =	tcstab->hasAttribute(TCSFIXED)
					? tcstab->attribute(TCSFIXED)
					: tcstab->addAttribute(TCSFIXED,
										   "TCS has been fixed by tcsfix"),
		*tcsfixv =	tcstab->hasAttribute(TCSFIXV)
					? tcstab->attribute(TCSFIXV)
					: tcstab->addAttribute(TCSFIXV,
									   "tcsfix version that fixed this TCS");

	*tcsfixed = true;
	*tcsfixv = version;
	*tcsfixcode = fixCode + tcsfixer.fixedCode();

	dataSetServer->close(tcstab->parent());
}
