//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	xmmtrackinfo --- ODF access functions for xmmtrack application (IDL)
//
//	DESCRIPTION:
//		This module contains a collection of SAS routines (mostly from OAL)
//		which are needed by the IDL application 'xmmtrack'. All functions
//		are tailored toward IDL's CALL_EXTERNAL I/F which strongly limits their
//		usefulness in other contexts.
//
//	CHANGE HISTORY:
//		UL, 2001-01-02: 1.0.0: - initial version
//		UL, 2001-12-04: 1.0.1: - siderial -> sidereal
//
//	AUTHOR:
//		U. Lammers, Tue Jan  2 16:51:40 MET 2001
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/tools/src/xmmtrack/xmmtrackinfo.cc,v 1.6 2001/12/04 17:10:35 ulammers Exp $
//
#include <string>

#include "STime.h"
#include "OalServer.h"
#include "Xmm.h"
#include "Celestial.h"
#include "SasFileLocator.h"
#include "stringutil.h"
#include "errstr.h"
#include "Tracer.h"

//
//	stuff from JPL ephemeris lib
//
int baryinit (int ephnum, const char *path);
const double *dpleph (double *jd, int ntarg, int ncent);

extern "C" {
	//
	//	ODF inquiry routines
	//

	//	revolution number
	int revolution(int, void *)
	{
		PROC
		RETURN oalServer.odf()->orbitNr();
	}


	//	observation start time
	double startTime(int, void *)
	{
		PROC
		RETURN oalServer.odf()->observationStartTime() -
					Xmm::missionReferenceTime;
	}


	//	observation end time
	double endTime(int, void *)
	{
		PROC
		RETURN oalServer.odf()->observationEndTime() -
					Xmm::missionReferenceTime;
	}


	//	convert MET tag (double) into absolute time (string)
	const char *absTime(int argc, void *argv[])
	{
		PROC
		static char timeBuff[32];

		if (argc != 1)
			errstr<<"Usage: absTime(time)"
				  <<Sas::error("WrongNumberOfArguments");

		double time = *(double *)argv[0];
		STime t = Xmm::missionReferenceTime + time;
		strncpy(timeBuff, t.toString(0).c_str(), sizeof(timeBuff));

		RETURN timeBuff;
	}


	//	compute Greenwich sidereal time corresponding to given MET tag
	double siderealTime(int argc, void *argv[])
	{
		PROC

		if (argc != 1)
			errstr<<"Usage: siderealTime(time)"
				  <<Sas::error("WrongNumberOfArguments");

		double time = *(double *)argv[0];
		STime t = Xmm::missionReferenceTime + time;

		RETURN t.siderealTime();
	}


	//	convert right ascension value (double) to string
	const char *raString(int argc, void *argv[])
	{
		PROC
		static char raBuff[32];

		if (argc != 1)
			errstr<<"Usage: raString(ra)"
				  <<Sas::error("WrongNumberOfArguments");

		double ra = *(double *)argv[0];
		strncpy(raBuff, RightAscension(ra).toString().c_str(), sizeof(raBuff));

		RETURN raBuff;
	}


	//	convert declination value (double) to string
	const char *decString(int argc, void *argv[])
	{
		PROC
		static char decBuff[32];

		if (argc != 1)
			errstr<<"Usage: decString(dec)"
				  <<Sas::error("WrongNumberOfArguments");

		double dec = *(double *)argv[0];
		strncpy(decBuff, Declination(dec).toString().c_str(), sizeof(decBuff));

		RETURN decBuff;
	}


	//	return position vector of spacecraft at given time (MET tag)
	typedef struct { double x, y, z; } PositionVector;
	int xmmPosition(int argc, void *argv[])
	{
		PROC

		static const PositionLocator *posLocator = oalServer.positionLocator();

		if (argc != 2)
			errstr<<"Usage: xmmPosition(time, posVec)"
				  <<Sas::error("WrongNumberOfArguments");

		double time = *(double *)argv[0];
		PositionVector *posVec = (PositionVector *)argv[1];

		try {
			CartesianVector pos = posLocator->stateVector(time).position;
			posVec->x = pos.x();
			posVec->y = pos.y();
			posVec->z = pos.z();
			ALTRETURNV(0);
		} catch(...) {}

		RETURN 1;
	}


	static XmmOrbitSimulator *orbitSim = 0;

	//	select orbital simulator with orbital elements valid at given
	//	absolute time (string)
	typedef struct { unsigned short slen;
					 short type;
					 char *s; } IDL_STRING;
	typedef struct { double incl,							//	[deg]
							raAscendingNode,				//	[deg]
							argPerigee; } OrbitElements;	//	[deg]
	int xmmPredictedOrbitAt(int argc, void *argv[])
	{
		PROC

		if (argc != 2)
			errstr<<"Usage: met(timeString, orbElements)"
				  <<Sas::error("WrongNumberOfArguments");

		char *timeString = ((IDL_STRING *)argv[0])->s;
		//cout<<"|"<<timeString<<"|"<<endl;
		if (STime::isValidTime(timeString)) {
			STime t(trim(timeString, true));

			delete orbitSim;
			orbitSim = new XmmOrbitSimulator(t);

			const OrbitalElements &oe = orbitSim->orbitalElements();
			OrbitElements *orbEl = (OrbitElements *)argv[1];
			orbEl->incl = oe.inclincation * RAD_TO_DEG;
			orbEl->raAscendingNode = oe.ascendingNode * RAD_TO_DEG;
			orbEl->argPerigee = oe.perigeeArgument * RAD_TO_DEG;

			ALTRETURNV(0);
		}

		RETURN 1;
	}


	//	compute predicted satellite positon vector at given time
	int xmmPredictedPosition(int argc, void *argv[])
	{
		PROC

		if (!orbitSim)
			orbitSim = new XmmOrbitSimulator(
									oalServer.odf()->observationStartTime());

		if (argc != 2)
			errstr<<"Usage: xmmPredictedPosition(time, posVec)"
				  <<Sas::error("WrongNumberOfArguments");

		double time = *(double *)argv[0];
		PositionVector *posVec = (PositionVector *)argv[1];

		CartesianVector pos = orbitSim->position(time);
		posVec->x = pos.x();
		posVec->y = pos.y();
		posVec->z = pos.z();

		RETURN 0;
	}


	//	compute position of sun at specific time
	int sunPosition(int argc, void *argv[])
	{
		PROC

		static bool jplInitialized = false;
		if (argc != 2)
			errstr<<"Usage: sunPosition(time, posVec)"
				  <<Sas::error("WrongNumberOfArguments");

		double time = *(double *)argv[0];
		PositionVector *posVec = (PositionVector *)argv[1];

		if (!jplInitialized) {
			//	initialize JPL ephemeris module

			string jplPath = fileLocator.locate("lib/barycendata");
			baryinit(200, jplPath.c_str());

			jplInitialized = true;
		}

		//	obtain time in MJD (TT)
		STime t = Xmm::missionReferenceTime + time;
		STime::JulianDayNumber jdn(t.modifiedJulianDay(STime::TT));
		double jdt[2] = { jdn.day()+2400001., jdn.fractionalDay()-.5 };

		static const int EARTH = 3,
						 SUN = 11;
		const double *eposn = dpleph(jdt, EARTH, SUN);
		if (!eposn) {
			errstr<<"Could not find position of sun at T="
				  <<t.toString()<<Sas::error("UnableToLocateSun");
			ALTRETURNV(1);
		}

		const double *earthPos = eposn,
					 *sunPos = eposn+6;
		CartesianVector earthSun(sunPos[0]-earthPos[0],
								 sunPos[1]-earthPos[1],
								 sunPos[2]-earthPos[2]);
		earthSun.normalize();
		posVec->x = earthSun.x();
		posVec->y = earthSun.y();
		posVec->z = earthSun.z();

		RETURN 0;
	}


	//	inquite spacecraft attitude at given time (in MET tag)
	typedef struct { double phi, theta, psi; } AngleVector;
	int xmmAttitude(int argc, void *argv[])
	{
		static const AttitudeLocator *attLocator = oalServer.attitudeLocator();

		if (argc != 2)
			errstr<<"Usage: xmmAttitude(time, attVec)"
				  <<Sas::error("WrongNumberOfArguments");

		double time = *(double *)argv[0];
		AngleVector *angVec = (AngleVector *)argv[1];

		try {
			EulerAngles euler = attLocator->attitude(time).toEulerAngles();
			angVec->phi = euler.phi();
			angVec->theta = euler.theta();
			angVec->psi = euler.psi();
			ALTRETURNV(0);
		} catch(...) {}

		RETURN 1;
	}
}

#if 0
void tryt(const string &st)
{
	STime abst(st);
	PositionVector v;
	double t = abst - Xmm::missionReferenceTime;
	void *argv[] = { &t, &v };
	sunPosition(2, argv);
	cout<<v.x<<endl;
	cout<<v.y<<endl;
	cout<<v.z<<endl<<endl;
}
#endif

#if 0
	tryt("2000-03-21T00:00:00");
	tryt("2000-06-21T00:00:00");
	tryt("2000-09-23T00:00:00");
	tryt("2000-12-21T00:00:00");
#endif
